# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
from mimesis import locales as Locales
from mimesis import Text

class Color:

    @classmethod
    def color(cls, *, locale=Locales.EN) -> str:
        '''
            Generate a random color string.

            Keyword Arguments:
                locale: (Optional) locale for generating street name

            Returns:
                A generated color
        '''
        return Text(locale).color()

    @classmethod
    def hex_color(cls) -> str:
        '''
            Generate a hex color code.

            Returns:
                A generated hex color code string
        '''
        return Text().hex_color()

    @classmethod
    def rgb_color(cls) -> tuple:
        '''
            Generate an rgb color tuple.

            Returns:
                RGB tuple
        '''
        return Text().rgb_color()




# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from arjuna.tpi.parser.yaml import YamlDict
from arjuna.tpi.data.generator import generator
import os

from arjuna.tpi.httpauto.response import HttpResponse
from arjuna.tpi.error import SEAMfulActionFileError
from arjuna.tpi.helper.arjtype import CIStringDict
from arjuna.tpi.data.entity import _DataEntity
from arjuna import Random
from arjuna.tpi.helper.arjtype import NotFound
from arjuna.tpi.parser.text import Text

def check_data_arg_type(kwargs):
    if 'data' in kwargs:
        if type(kwargs['data']) is not dict and not isinstance(kwargs['data'], _DataEntity):
            raise Exception("'data' keyword argument for action call or action.perform() call can only be a Python dict or an Arjuna Data Entity. Provided: >>{}<< of type >>{}<<".format(kwargs['data'], type(kwargs['data'])))


class _HttpActionStep:

    def __init__(self, action, step_yaml):
        self.__action = action
        self.__msg = step_yaml

    @property
    def _message(self):
        return self.__msg

    def perform(self, **fargs):
        return self.__action.send(self.__msg, **fargs)

class _HttpActionSteps:

    def __init__(self, action, steps_yaml=None):
        self.__action = action
        self.__steps = steps_yaml is not None and steps_yaml or []

    def __iter__(self):
        return iter([_HttpActionStep(self.__action, step_yaml) for step_yaml in self.__steps])

class BaseHttpEndPointAction:

    def __init__(self, *, name, endpoint):
        self.__name = name
        self.__endpoint = endpoint
        from .msgloader import HttpMessageLoader
        self._messages = HttpMessageLoader(action=self)

    @property
    def name(self):
        return self.__name

    @property
    def _message(self):
        return self._messages

    @property
    def _endpoint(self):
        return self.__endpoint

    def send(self, msg=None, **fargs) -> HttpResponse:
        check_data_arg_type(fargs)
        return self._message.send(msg, **fargs)

    def perform(self, **fargs) -> HttpResponse:
        pass

    def _get_file_path_msg(self):
        return ""

    def __str__(self):
        meta_str = f"HTTP Action: >{self.name}<" + self._get_file_path_msg()
        if self._endpoint.name != "anon":
            meta_str += f" for end point >{self._endpoint.name}<"
        if self._endpoint.service.name != "anon":
            meta_str += f" for service >{self._endpoint.service.name}<"
        return meta_str

    __repr__ = __str__

class AnonEndPointAction(BaseHttpEndPointAction):

    def __init__(self, *, endpoint):
        super().__init__(name="anon", endpoint=endpoint)

    def perform(self, **fargs):
        pass

class HttpMessageResponse:

    def __init__(self, *, step, msg, response):
        self.__step = step
        self.__msg = msg
        self.__response = response

    @property
    def step(self):
        '''
            Step number for message in the action starting from 1.
        '''
        return self.__step

    @property
    def msg(self):
        '''
           HTTP Message name.
        '''
        return self.__msg

    @property
    def response(self):
        '''
            HTTP Response object.
        '''
        return self.__response

    def __str__(self):
        return str({
            "step": self.step,
            "msg": self.msg,
            "response": self.response
        })

    __repr__ = __str__

class HttpEndPointAction(BaseHttpEndPointAction):

    def __init__(self, *, name, endpoint, **fargs):
        super().__init__(name=name, endpoint=endpoint)
        check_data_arg_type(fargs)
        self.__fargs = fargs
        # Process Yaml file
        from arjuna import C, Yaml
        # Check Built-In Security Action
        my_dir = os.path.dirname(os.path.realpath(__file__))
        self.__action_file_path = os.path.abspath(os.path.join(my_dir, "..", "..","..", "res", "security", "http", "action", name + ".yaml"))
        try:
            f = open(self.__action_file_path, "r")
        except FileNotFoundError:
            self.__action_file_path = os.path.join(endpoint.root_dir, "action", name + ".yaml")
            try:
                f = open(self.__action_file_path, "r")
            except FileNotFoundError:
                raise SEAMfulActionFileError(self, msg=f"Action file not found at location: {self.__action_file_path}")
        self.__msg_yaml = f.read()
        f.close()
        self.__store = CIStringDict()

    def _get_file_path_msg(self):
        return f" at ({self.__action_file_path})"

    @property
    def store(self):
        '''
        Values extracted from response and stored.
        '''
        return self.__store

    def _load(self, **sfargs):
        from arjuna import C, Yaml
        margs = {}
        margs.update(self.__fargs)
        margs.update(sfargs)
        from arjuna.core.fmt import arj_convert
        margs = {k:arj_convert(v) for k,v in margs.items()}
        action_yaml = Text(self.__msg_yaml).format(**margs)
        action_yaml = Yaml.from_str(action_yaml, allow_any=True)
        if action_yaml is None:
            return CIStringDict(), margs, list()

        def call_generator(in_dict):
            d = {}
            d.update(in_dict)
            gen = d["generator"]
            del d["generator"]
            return getattr(Random, gen)(**d)       

        def gen_or_value(value):
            if isinstance(value, YamlDict):
                if "generator" in value:
                    return call_generator(value)
                else:
                    return value
            else:
                return value

        if 'data' not in self.store:
            self.store['data'] = CIStringDict()

        if "data" in action_yaml:
            for name, instruction in action_yaml["data"].items():
                if type(instruction) is dict or isinstance(instruction, YamlDict):
                    if "generator" in instruction:
                        self.store['data'][name] = call_generator(instruction)
                    else:
                        self.store['data'][name] = instruction
            
        if "entity" in action_yaml:
            from arjuna.core.importer import import_arj_entity
            for name, instruction in action_yaml["entity"].items():
                if name == "data":
                    raise SEAMfulActionFileError(self, msg=f"Entity name can not be >>data<< as it is reserved space for action.store.data. Validate action file: {self.__action_file_path}")                 
                if type(instruction) is str:
                    entity = instruction
                    kwargs = {}
                elif type(instruction) is dict or isinstance(instruction, YamlDict):
                    entity = instruction['type']
                    kwargs = {}
                    kwargs.update(instruction.items())
                    del kwargs['type']
                else:
                    raise SEAMfulActionFileError(self, msg=f"Entity >>{entity}<< can not be loaded. Its value can either be the Entity Class Name or dictionary has 'type' key set to Entity Class Name and rest of them are keyword arguments. Validate action file: {self.__action_file_path}")                 
                try:
                    entity_callable = import_arj_entity(entity)
                except Exception as e:
                    raise SEAMfulActionFileError(self, msg=f"Entity >>{entity}<< can not be loaded. Error: {e}. Validate action file: {self.__action_file_path}")                 
                else:
                    entity_kwarg_dict = {k:gen_or_value(v) for k,v in kwargs.items()}
                    self.store[name] = entity_callable(**entity_kwarg_dict)

        if "alias" in action_yaml:
            for k, v in action_yaml["alias"].items():
                if v in self.store:
                    self.store[k] = self.store[v]
                elif v in self.store['data']:
                    self.store[k] = self.store['data'][v]
                elif v.split(".")[0] in self.store:
                    ename, attr = v.split(".", 1) 
                    if isinstance(self.store[ename], _DataEntity):
                        self.store[k] = getattr(self.store[ename], attr)
                else:
                    self.store[k] = v
        
        if 'data' in margs:
            if isinstance(margs['data'], _DataEntity):
                margs['data'] = margs['data'].as_dict()
        else:
            margs['data'] = dict()
        margs['data'].update(self.store['data'])

        margs.update({k:v for k,v in self.store.items() if k != "data"})

        mproc = None
        if "messages" in action_yaml:
            mproc = _HttpActionSteps(self, action_yaml['messages'])
            del action_yaml["messages"]
        else:
            mproc = _HttpActionSteps(self)
        
        return CIStringDict(), margs, mproc

    def perform(self, **fargs):
        from arjuna import log_info
        log_info(f"Performing", self)
        check_data_arg_type(fargs)
        responses = list()
        meta, fargs, steps = self._load(**fargs)
        for index, step in enumerate(steps):
            response = step.perform(**fargs)
            new_data = {k:v for k,v in response.store.items() if k != "default_content" and not isinstance(v, NotFound)}
            self.store.update(new_data)
            fargs.update(new_data)
            responses.append(HttpMessageResponse(step=index+1, msg=step._message, response=response))
        return responses



# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from arjuna.tpi.httpauto.response import HttpResponse
from arjuna.interact.http.session import HttpSession
from .seamful.eploader import HttpEndPointLoader

class HttpService:
    '''
        HTTP Service representing SEAMful automation in Arjuna.

        Keyword Arguments:
            name: Name of service. Should have a corresponding named directory in Project root/httpauto/service directory. If not provided then the name is set to **anon** and root directory is set to Project root/httpauto.
            url: Base URL for this HTTP session. If relative path is used as a route in sender methods like `.get`, then this URL is prefixed to their provided routes.
            oauth_token: OAuth 2.0 Bearer token for this service.
            request_content_handler: Default content type handler for requests sent in this service. Overridable in individual sender methods. Default is Http.content.urlencoded.
            headers: HTTP headers to be added to request headers made by this service.
            max_redirects: Maximum number of redirects allowed for a request. Default is 30.
            auth: HTTP Authentication object: Basic/Digest.
            proxy: Proxies dict to be associated with this service.
    '''

    def __init__(self, *, name="anon", url=None, oauth_token=None, request_content_handler=None, headers=None, max_redirects=None, auth=None, proxy=None, _auto_session=True):
        self.__name = name
        self.__session = HttpSession(url=url, oauth_token=oauth_token, request_content_handler=request_content_handler, headers=headers, max_redirects=max_redirects, auth=auth, proxy=proxy, _auto_session=_auto_session)
        from arjuna import C
        if name == "anon":
            self.__root_dir = C("httpauto.dir")
        else:
            self.__root_dir = os.path.join(C("httpauto.dir"), "service/{}".format(name))
        self.__endpoints = HttpEndPointLoader(self)

    def _set_session(self, session):
        self.__session._set_session(session)

    @property
    def root_dir(self):
        return self.__root_dir

    @property
    def name(self):
        '''
            Name of service
        '''
        return self.__name

    @property
    def _session(self):
        '''
            HTTP Session object associated with this service.
        '''
        return self.__session

    @property
    def ep(self):
        '''
            Http End Point loader for this service.
        '''
        return self.__endpoints

    @property
    def message(self):
        '''
            Http Message Loader for this service using default end point.
        '''
        return self.ep._anon.message

    @property
    def action(self):
        '''
            Http Action Loader for this service using default end point.
        '''
        return self.ep._anon.action

    def send(self, msg=None, **fargs) -> HttpResponse:
        '''
            Send an HTTP Message using this service using default end point.
        '''
        return self.ep._anon.send(msg=msg, **fargs)
    
    def perform(self, action=None, **fargs) -> HttpResponse:
        '''
            Perform an HTTP Action using this service using default end point.
        '''
        return self.ep._anon.perform(action=action, **fargs)

    @property
    def cookies(self) -> dict:
        '''
            All current cookie name/values in this service.
        '''
        return self._session.cookies

    @property
    def parsed_cookies(self) -> dict:
        '''
            All current cookie name and corresponding `HttpCookie` object in this service.
        '''
        return self._session.parsed_cookies

    def add_cookies(self, cookie_dict):
        '''
            Add cookies to the service.
        '''
        self._session.add_cookies(cookie_dict)

    @property
    def request_content_handler(self):
        '''
        Request content handler for content formatting.
        '''
        return self.__request_content_handler

    @property
    def headers(self):
        '''
        Request headers.
        '''
        return self._session.headers

    @property
    def auth(self):
        '''
        HTTP Authentication object.
        '''
        return self._session.auth

    @property
    def url(self):
        '''
            Base URL for this session.
        '''
        return self._session.url

    @property
    def _session(self):
        return self.__session

    @property
    def _request_headers(self):
        return self.__session._request_headers

    def get(self, route, label=None, xcodes=None, headers=None, cookies=None, allow_redirects=True, auth=None, timeout: float=None, pretty_url=False, query_params=None, **named_query_params) -> HttpResponse:
        '''
        Sends an HTTP GET request.

        Arguments:
            route: Absolute or relative URL. If relative, then `url` of this session object is pre-fixed.

        Keyword Arguments:
            label: Label for this request. If available, it is used in reports and logs.
            xcodes: Expected HTTP response code(s).
            headers: Mapping of additional HTTP headers to be sent with this request.
            cookies: Python dict of cookies to send with request.
            allow_redirects: If True, redirections are allowed for the HTTP message. Default is True.
            auth: HTTP Authentication object: Basic/Digest.
            timeout: How long to wait for the server to send data before giving up.
            pretty_url: If True, the query params are formatted using pretty URL format instead of usual query string which is the default.
            query_params: A mapping of key-values to be included in query string.
            **named_query_params: Arbitrary key/value pairs. These are appended to the query string of URL for this request.

        Note:
            **query_params** and **named_query_params** have the same goal.
            In case of duplicates, named_query_params override query_params.
        '''
        return self._session.get(route, label=label, xcodes=xcodes, headers=headers, cookies=cookies, allow_redirects=allow_redirects, auth=auth, timeout=timeout, pretty_url=pretty_url, query_params=query_params, **named_query_params)

    def head(self, route, label=None, xcodes=None, headers=None, cookies=None, allow_redirects=True, auth=None, timeout: float=None, pretty_url=False, query_params=None, **named_query_params) -> HttpResponse:
        '''
        Sends an HTTP HEAD request.

        Arguments:
            route: Absolute or relative URL. If relative, then `url` of this session object is pre-fixed.

        Keyword Arguments:
            label: Label for this request. If available, it is used in reports and logs.
            xcodes: Expected HTTP response code(s).
            headers: Mapping of additional HTTP headers to be sent with this request.
            cookies: Python dict of cookies to send with request.
            allow_redirects: If True, redirections are allowed for the HTTP message. Default is True.
            auth: HTTP Authentication object: Basic/Digest.
            timeout: How long to wait for the server to send data before giving up.
            pretty_url: If True, the query params are formatted using pretty URL format instead of usual query string which is the default.
            query_params: A mapping of key-values to be included in query string.
            **named_query_params: Arbitrary key/value pairs. These are appended to the query string of URL for this request.

        Note:
            **query_params** and **named_query_params** have the same goal.
            In case of duplicates, named_query_params override query_params.
        '''
        return self._session.head(route, label=label, xcodes=xcodes, headers=headers, cookies=cookies, allow_redirects=allow_redirects, auth=auth, timeout=timeout, pretty_url=pretty_url, query_params=query_params, **named_query_params)

    def delete(self, route, label=None, xcodes=None, headers=None, cookies=None, allow_redirects=True, auth=None, timeout: float=None, pretty_url=False, query_params=None, **named_query_params) -> HttpResponse:
        '''
        Sends an HTTP DELETE request.

        Arguments:
            route: Absolute or relative URL. If relative, then `url` of this session object is pre-fixed.

        Keyword Arguments:
            label: Label for this request. If available, it is used in reports and logs.
            xcodes: Expected HTTP response code(s).
            headers: Mapping of additional HTTP headers to be sent with this request.
            cookies: Python dict of cookies to send with request.
            allow_redirects: If True, redirections are allowed for the HTTP message. Default is True.
            auth: HTTP Authentication object: Basic/Digest.
            timeout: How long to wait for the server to send data before giving up.
            pretty_url: If True, the query params are formatted using pretty URL format instead of usual query string which is the default.
            query_params: A mapping of key-values to be included in query string.
            **named_query_params: Arbitrary key/value pairs. These are appended to the query string of URL for this request.

        Note:
            **query_params** and **named_query_params** have the same goal.
            In case of duplicates, named_query_params override query_params.
        '''
        return self._session.delete(route, label=label, xcodes=xcodes, headers=headers, cookies=cookies, allow_redirects=allow_redirects, auth=auth, timeout=timeout, pretty_url=pretty_url, query_params=query_params, **named_query_params)

    def post(self, route, *, content="", label=None, xcodes=None, headers=None, cookies=None, allow_redirects=True, auth=None, timeout: float=None, pretty_url=False, query_params=None, **named_query_params) -> HttpResponse:
        '''
        Sends an HTTP POST request.

        Arguments:
            route: Absolute or relative URL. If relative, then `url` of this session object is pre-fixed.

        Keyword Arguments:
            label: Label for this request. If available, it is used in reports and logs.
            content: Content to be sent in this HTTP request. If passed as string, then content-type set in session is used using the content request handler. It can also be a dictionary with keys - 'content' and 'type'. Default is empty string.
            xcodes: Expected HTTP response code(s).
            headers: Mapping of additional HTTP headers to be sent with this request.
            cookies: Python dict of cookies to send with request.
            allow_redirects: If True, redirections are allowed for the HTTP message. Default is True.
            auth: HTTP Authentication object: Basic/Digest.
            timeout: How long to wait for the server to send data before giving up.
            pretty_url: If True, the query params are formatted using pretty URL format instead of usual query string which is the default.
            query_params: A mapping of key-values to be included in query string.
            **named_query_params: Arbitrary key/value pairs. These are appended to the query string of URL for this request.

        Note:
            **query_params** and **named_query_params** have the same goal.
            In case of duplicates, named_query_params override query_params.
        '''
        return self._session.post(route, label=label, content=content, xcodes=xcodes, headers=headers, cookies=cookies, allow_redirects=allow_redirects, auth=auth, timeout=timeout, pretty_url=pretty_url, query_params=query_params, **named_query_params)

    def put(self, route, *, content="", label=None, xcodes=None, headers=None, cookies=None, allow_redirects=True, auth=None, timeout: float=None, pretty_url=False, query_params=None, **named_query_params) -> HttpResponse:
        '''
        Sends an HTTP PUT request.

        Arguments:
            route: Absolute or relative URL. If relative, then `url` of this session object is pre-fixed.

        Keyword Arguments:
            label: Label for this request. If available, it is used in reports and logs.
            content: Content to be sent in this HTTP request. If passed as string, then content-type set in session is used using the content request handler. It can also be a dictionary with keys - 'content' and 'type'. Default is empty string.
            xcodes: Expected HTTP response code(s).
            headers: Mapping of additional HTTP headers to be sent with this request.
            cookies: Python dict of cookies to send with request.
            allow_redirects: If True, redirections are allowed for the HTTP message. Default is True.
            auth: HTTP Authentication object: Basic/Digest.
            timeout: How long to wait for the server to send data before giving up.
            pretty_url: If True, the query params are formatted using pretty URL format instead of usual query string which is the default.
            query_params: A mapping of key-values to be included in query string.
            **named_query_params: Arbitrary key/value pairs. These are appended to the query string of URL for this request.

        Note:
            **query_params** and **named_query_params** have the same goal.
            In case of duplicates, named_query_params override query_params.
        '''
        return self._session.put(route, label=label, content=content, xcodes=xcodes, headers=headers, cookies=cookies, allow_redirects=allow_redirects, auth=auth, timeout=timeout, pretty_url=pretty_url, query_params=query_params, **named_query_params)

    def patch(self, route, *, content="", label=None, xcodes=None, headers=None, cookies=None, allow_redirects=True, auth=None, timeout: float=None, pretty_url=False, query_params=None, **named_query_params) -> HttpResponse:
        '''
        Sends an HTTP PATCH request.

        Arguments:
            route: Absolute or relative URL. If relative, then `url` of this session object is pre-fixed.

        Keyword Arguments:
            label: Label for this request. If available, it is used in reports and logs.
            content: Content to be sent in this HTTP request. If passed as string, then content-type set in session is used using the content request handler. It can also be a dictionary with keys - 'content' and 'type'. Default is empty string.
            xcodes: Expected HTTP response code(s).
            headers: Mapping of additional HTTP headers to be sent with this request.
            cookies: Python dict of cookies to send with request.
            allow_redirects: If True, redirections are allowed for the HTTP message. Default is True.
            auth: HTTP Authentication object: Basic/Digest.
            timeout: How long to wait for the server to send data before giving up.
            pretty_url: If True, the query params are formatted using pretty URL format instead of usual query string which is the default.
            query_params: A mapping of key-values to be included in query string.
            **named_query_params: Arbitrary key/value pairs. These are appended to the query string of URL for this request.

        Note:
            **query_params** and **named_query_params** have the same goal.
            In case of duplicates, named_query_params override query_params.
        '''
        return self._session.patch(route, label=label, content=content, xcodes=xcodes, headers=headers, cookies=cookies, allow_redirects=allow_redirects, auth=auth, timeout=timeout, pretty_url=pretty_url, query_params=query_params, **named_query_params)

    def options(self, route, *, label=None, xcodes=None, headers=None, cookies=None, allow_redirects=True, auth=None, timeout: float=None, pretty_url=False, query_params=None, **named_query_params) -> HttpResponse:
        '''
        Sends an HTTP OPTIONS request.

        Arguments:
            route: Absolute or relative URL. If relative, then `url` of this session object is pre-fixed.

        Keyword Arguments:
            label: Label for this request. If available, it is used in reports and logs.
            xcodes: Expected HTTP response code(s).
            headers: Mapping of additional HTTP headers to be sent with this request.
            cookies: Python dict of cookies to send with request.
            allow_redirects: If True, redirections are allowed for the HTTP message. Default is True.
            auth: HTTP Authentication object: Basic/Digest.
            timeout: How long to wait for the server to send data before giving up.
            pretty_url: If True, the query params are formatted using pretty URL format instead of usual query string which is the default.
            query_params: A mapping of key-values to be included in query string.
            **named_query_params: Arbitrary key/value pairs. These are appended to the query string of URL for this request.

        Note:
            **query_params** and **named_query_params** have the same goal.
            In case of duplicates, named_query_params override query_params.
        '''
        return self._session.options(route, label=label, content=content, xcodes=xcodes, headers=headers, cookies=cookies, allow_redirects=allow_redirects, auth=auth, timeout=timeout, pretty_url=pretty_url, query_params=query_params, **named_query_params)



# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from mimesis import locales as Locales
from mimesis import Person
from mimesis import Address

class Entity:

    @classmethod
    def address(cls, *, 
            locale=Locales.EN,
            calling_code=None, 
            city=None, 
            country=None, 
            country_code=None,
            latitude=None, 
            longitude=None,
            postal_code=None, 
            state=None,
            street_name=None,
            street_number=None,
            street_suffix=None
        ):
        '''
            Create an Address Data Entity object.

            All individual fields are automatically randomly generated based on locale. If provided, the corresponding values are overriden.

            Note:
                All individual fields are randomly generated. Don't expect correct correlation e.g. correct postal code for the generated city.

            Keyword Arguments:
                locale: Approprite Random.locale.<local_name> object. Default is Random.locale.EN
                calling_code: Calling Code
                city: City
                country: Country Name
                country_code: Country Code 
                latitude: Latitude
                longitude: Longitde
                postal_code: Postal Code
                state: State
                street_name: Street Name
                street_number Street Number
                street_suffix: Street Suffix
        '''
        address = Address(locale=locale)
        from arjuna.engine.data.entity.address import Address as ArjAddress

        return ArjAddress(
            calling_code=calling_code is not None and calling_code or address.calling_code(), 
            city=city and city is not None or address.city(), 
            country=country is not None and country or address.country(), 
            country_code=country_code is not None and country_code or address.country_code(), 
            latitude=latitude is not None and latitude or address.latitude(), 
            longitude=longitude is not None and longitude or address.longitude(),
            postal_code=postal_code is not None and postal_code or address.postal_code(), 
            state=state is not None and state or address.state(),
            street_name=street_name is not None and street_name or address.street_name(),
            street_number=street_number is not None and street_number or address.street_number(),
            street_suffix=street_suffix is not None and street_suffix or address.street_suffix(),
        )

    @classmethod
    def person(cls, *, 
            locale=Locales.EN,
            qualification=None,
            age=None,
            blood_type=None,
            email=None,
            first_name=None,
            last_name=None,
            gender=None,
            height=None,
            id=None,
            language=None,
            nationality=None,
            occupation=None,
            phone=None,
            title=None,
            university=None,
            weight=None,
            work_experience=None,

        ):
        '''
            Create an Person Data Entity object.

            All individual fields are automatically randomly generated based on locale. If provided, the corresponding values are overriden.

            Note:
                All individual fields are randomly generated. Don't expect correct correlation e.g. correct postal code for the generated city.

            Keyword Arguments:
                locale: Approprite Random.locale.<local_name> object. Default is Random.locale.EN
                qualification: Educational Qualification
                age: Age
                blood_type: Blood type
                email: Email address
                first_name: First name
                last_name: Last name
                gender: Gender
                height: Height
                id: Identifier
                language: Language
                nationality: Nationality
                occupation: Occupation
                phone: Phone number
                title: Title
                university: University
                weight: Weight
                work_experience: Work Experience
        '''
        person = Person(locale=locale)
        from arjuna.engine.data.entity.person import Person as ArjPerson

        first_name=first_name is not None and first_name or person.first_name()
        last_name=last_name is not None and last_name or person.last_name()
        return ArjPerson(
            qualification=qualification is not None and qualification or person.academic_degree(),
            age=age is not None and age or person.age(),
            blood_type=blood_type is not None and blood_type or person.blood_type(),
            email=email is not None and email or person.email(),
            first_name=first_name,
            last_name=last_name,
            name=first_name + " " + last_name,
            gender=gender is not None and gender or person.gender(),
            height=height is not None and height or person.height(),
            id=id is not None and id or person.identifier(),
            language=language is not None and language or person.language(),
            nationality=nationality is not None and nationality or person.nationality(),
            occupation=occupation is not None and occupation or person.occupation(),
            phone=phone is not None and phone or person.telephone(),
            title=title is not None and title or person.title(),
            university=university is not None and university or person.university(),
            weight=weight is not None and weight or person.weight(),
            work_experience=work_experience is not None and work_experience or person.work_experience(),
        )





# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib

def import_module(mod_name, *, prefix="", optional=False):
    full_path = mod_name
    if prefix: 
        ".".join([prefix, full_path])
    try:
        mod = importlib.import_module(full_path)
    except ModuleNotFoundError:
        if not optional:
            raise
    else:
        return mod 

def import_name_in_module(*, mod_name, name, prefix="", optional=False):
    mod = import_module(mod_name, prefix=prefix, optional=optional)
    if mod is None:
        return None

    try:
        return getattr(mod, name)
    except AttributeError:
        if not optional:
            raise ImportError("{} is not defined in module: {}".format(name, mod))

def import_arj_hook_module(mod_name, optional=False):
    from arjuna import C
    return import_module(mod_name, prefix=C("hooks.package"), optional=optional)

def import_name_in_arj_hook_module(*, mod_name, name, optional=False):
    from arjuna import C
    mod = import_arj_hook_module(mod_name)
    return import_name_in_module(mod_name=mod_name, name=name, prefix=C("hooks.package"), optional=optional)

def import_arj_resouce_hook_package():
    from arjuna import C
    import_name_in_module(mod_name=C("hooks.config.package"), name='*', optional=True)

def import_arj_config_hook(name):
    from arjuna import C
    return import_name_in_module(mod_name=C("hooks.config.package"), name=name, optional=True)

def import_arj_entity(name):
    from arjuna import C
    return import_name_in_module(mod_name=C("hooks.entity.package"), name=name, optional=False)


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from arjuna.interact.http.model.internal.repr.request import HttpRequestYamlRepr
from arjuna.interact.http.model.internal.repr.response import HttpResponseYamlRepr
from arjuna.tpi.helper.arjtype import CIStringDict
from arjuna.tpi.error import SEAMfulMessageFileError
from arjuna.tpi.parser.text import Text

def check_data_arg_type(kwargs):
    from arjuna.tpi.data.entity import _DataEntity
    if 'data' in kwargs:
        if type(kwargs['data']) is not dict and not isinstance(kwargs['data'], _DataEntity):
            raise Exception("'data' keyword argument for HTTP message call or HTTP message.send() call can only be a Python dict or an Arjuna Data Entity. Provided: >>{}<< of type >>{}<<".format(kwargs['data'], type(kwargs['data'])))


class BaseHttpMessage:

    def __init__(self, name, action):
        self.__name = name
        self.__action = action

    @property
    def name(self):
        return self.__name

    @property
    def _action(self):
        return self.__action

    def send(self, **fargs):
        from arjuna import log_info
        log_info("Sending", self)
        check_data_arg_type(fargs)
        label, req_repr, resp_proc = self._load(**fargs)
        method = req_repr.method
        try:
            call = getattr(self._action._endpoint.service._session, method)
        except AttributeError:
            raise Exception(f"Unsupported HTTP method: {method}")
        else:
            response = call(req_repr.route, label=label, **req_repr.attrs)
            resp_proc.validate(response)
            return response

    def _load(self, **sfargs):
        pass

    def _get_file_path_msg(self):
        return ""

    def __str__(self):
        meta_str = f"HTTP Message >{self.name}<" + self._get_file_path_msg()
        if self._action.name != "anon":
            meta_str += f" for action >{self._action.name}<"
        if self._action._endpoint.name != "anon":
            meta_str += f" for end point >{self._action._endpoint.name}<"
        if self._action._endpoint.service.name != "anon":
            meta_str += f" for service >{self._action._endpoint.service.name}<"
        return meta_str

    __repr__ = __str__


class RootHttpMessage(BaseHttpMessage):

    def __init__(self, action):
        super().__init__(name="Root", action=action)

    def _load(self, **sfargs):
        req_repr = HttpRequestYamlRepr(self._action, CIStringDict(), label="Root")
        resp_proc = HttpResponseYamlRepr(self._action, CIStringDict())
        return "Root", req_repr, resp_proc

class HttpMessage(BaseHttpMessage):

    def __init__(self, *, name, action, **fargs):
        super().__init__(name=name, action=action, **fargs)
        check_data_arg_type(fargs)
        self.__fargs = fargs
        # Process Yaml file
        from arjuna import C, Yaml
        # Check Built-In Security Message
        my_dir = os.path.dirname(os.path.realpath(__file__))
        self.__action_file_path = os.path.abspath(os.path.join(my_dir, "..", "..","..", "res", "security", "http", "message", name + ".yaml"))
        try:
            f = open(self.__action_file_path, "r")
        except FileNotFoundError:
            self.__file_path = os.path.join(action._endpoint.root_dir, "message", name + ".yaml")
            try:
                f = open(self.__file_path, "r")
            except FileNotFoundError:
                raise SEAMfulMessageFileError(self, msg=f"Message file not found at location: {self.__file_path}")
        self.__msg_yaml = f.read()
        f.close()
        super().__init__(name=name, action=action)

    def _get_file_path_msg(self):
        return f" at ({self.__file_path})"

    def _load(self, **sfargs):
        from arjuna import C, Yaml
        margs = {}
        margs.update(self.__fargs)
        margs.update(sfargs)
        from arjuna.core.fmt import arj_convert
        margs = {k:arj_convert(v) for k,v in margs.items()}
        msg_yaml = Text(self.__msg_yaml).format(**margs)
        msg_yaml = Yaml.from_str(msg_yaml, allow_any=True)
        if msg_yaml is None:
            req_repr = HttpRequestYamlRepr(self._action, CIStringDict(), label="Root")
            resp_proc = HttpResponseYamlRepr(self._action, CIStringDict())
            return self.name, req_repr, resp_proc

        if "label" in msg_yaml:
            label = msg_yaml["label"]
            del msg_yaml["label"]
        else: 
            label = self.name

        if "request" in msg_yaml:
            req_repr = HttpRequestYamlRepr(self._action, CIStringDict(msg_yaml["request"].as_map()), label=label)
            del msg_yaml["request"]
        else:
            req_repr = HttpRequestYamlRepr(self._action, CIStringDict(), label=label)
        
        resp_proc = HttpResponseYamlRepr(self._action, CIStringDict(msg_yaml.as_map()))
        return label, req_repr, resp_proc











# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .message import RootHttpMessage, HttpMessage

class HttpMessageLoader:

    def __init__(self, action):
        self._action = action

    def __getattr__(self, msg_name: str, **fargs) -> HttpMessage:
        if msg_name == "root":
            msg = RootHttpMessage(vars(self)["_action"])
        else:
            msg = HttpMessage(name=msg_name, action=vars(self)["_action"], **fargs)

        return msg

    def send(self, msg=None, **fargs):
        if msg is None:
            return self.root.send(**fargs)
        else:
            return getattr(self, msg).send(**fargs)


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from mimesis import locales as Locales
from mimesis import Address as MimesisAddress

class Address:

    @classmethod
    def street_name(cls, *, locale=Locales.EN) -> str:
        '''
            Generate a street name

            Keyword Arguments:
                locale: (Optional) locale for generating street name

            Returns:
                A generated street name
        '''
        return MimesisAddress(locale).street_name()

    @classmethod
    def street_number(cls, *, locale=Locales.EN, prefix="") -> str:
        '''
            Generate a street number

            Keyword Arguments:
                locale: (Optional) locale for generating street number
                prefix: (Optional) prefix to be added to the generated street number. Space is used as delimited between prefix and generated street number.

            Returns:
                A generated street number
        '''
        if prefix != "":
            prefix = f"{prefix} "
        return prefix + MimesisAddress(locale).street_number()

    @classmethod
    def house_number(cls, *, locale=Locales.EN, prefix="") -> str:
        '''
            Generate a house number

            Keyword Arguments:
                locale: (Optional) locale for generating house number
                prefix: (Optional) prefix to be added to the generated house number. Space is used as delimited between prefix and generated house number.

            Returns:
                A generated house number
        '''
        return cls.street_number(locale=locale, prefix=prefix)

    @classmethod
    def postal_code(cls, *, locale=Locales.EN) -> str:
        '''
            Generate a postal code

            Keyword Arguments:
                locale: (Optional) locale for generating postal code

            Returns:
                A generated postal code
        '''
        return MimesisAddress(locale).postal_code()

    @classmethod
    def city(cls, *, locale=Locales.EN) -> str:
        '''
            Generate a city name.

            Keyword Arguments:
                locale: (Optional) locale for generating city name

            Returns:
                A generated city name
        '''
        return MimesisAddress(locale).city()

    @classmethod
    def country(cls, *, locale=Locales.EN) -> str:
        '''
            Generate a country name

            Keyword Arguments:
                locale: (Optional) locale for generating country name

            Returns:
                A generated country name
        '''
        return MimesisAddress(locale).country()



# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .action import *
from .endpoint import AnonHttpEndPoint, HttpEndPoint

class HttpEndPointLoader:

    def __init__(self, service):
        self._service = service
        self._anon_ep = None

    def __getattr__(self, ep_name: str, **fargs) -> HttpEndPointAction:
        if ep_name == "_anon":
            if vars(self)['_anon_ep'] is None:
                vars(self)['_anon_ep']  = AnonHttpEndPoint(service=vars(self)['_service'])
            mep = vars(self)['_anon_ep']
        else:
            mep = HttpEndPoint(vars(self)["_service"], name=ep_name, **fargs)
        return mep

    def send(self, **fargs):
        self._anon.send(**fargs)


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from arjuna.tpi.data.entity import data_entity

Address = data_entity(
        "Address", 
        calling_code=None, 
        city=None, 
        continent=None, 
        coordinates=None, 
        country=None, 
        country_code=None, 
        latitude=None, 
        longitude=None, 
        postal_code=None, 
        state=None,
        street_name=None,
        street_number=None,
        street_suffix=None,
    )


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This has to wait. Mimesis 5.0.0 release on Pypi not available yet.
# from mimesis import BinaryFile
#     from mimesis.enums import AudioFile, CompressedFile, DocumentFile, ImageFile, VideoFile

class File:

    @classmethod
    def mp3(cls) -> bytes:
        '''
            Generate an MP3 audio file and return as bytes.
        '''
        return BinaryFile().audio(file_type=AudioFile.MP3)

    @classmethod
    def aac(cls):
        '''
            Generate an AAC audio file and return as bytes.
        '''
        return BinaryFile().audio(file_type=AudioFile.ACC)

    @classmethod
    def zip(cls):
        '''
            Generate a zip compressed file and return as bytes.
        '''
        return BinaryFile().audio(file_type=CompressedFile.ZIP)

    @classmethod
    def gzip(cls):
        '''
            Generate a gzip compressed file and return as bytes.
        '''
        return BinaryFile().audio(file_type=CompressedFile.gzip)

    @classmethod
    def pdf(cls):
        '''
            Generate a PDF document file and return as bytes.
        '''
        return BinaryFile().compressed(file_type=DocumentFile.PDF)

    @classmethod
    def docx(cls):
        '''
            Generate a DOCX document file and return as bytes.
        '''
        return BinaryFile().compressed(file_type=DocumentFile.DOCX)

    @classmethod
    def pptx(cls):
        '''
            Generate a PPTX document file and return as bytes.
        '''
        return BinaryFile().document(file_type=DocumentFile.PPTX)

    @classmethod
    def xlsx(cls):
        '''
            Generate a XLSX document file and return as bytes.
        '''
        return BinaryFile().document(file_type=DocumentFile.XLSX)

    @classmethod
    def gif(cls):
        '''
            Generate a GIF Image file and return as bytes.
        '''
        return BinaryFile().image(file_type=ImageFile.GIF)

    @classmethod
    def png(cls):
        '''
            Generate a PNG Image file and return as bytes.
        '''
        return BinaryFile().image(file_type=ImageFile.PNG)

    @classmethod
    def jpg(cls):
        '''
            Generate a JPG Image file and return as bytes.
        '''
        return BinaryFile().image(file_type=ImageFile.JPG)

    @classmethod
    def mov(cls):
        '''
            Generate a MOV Video file and return as bytes.
        '''
        return BinaryFile().image(file_type=VideoFile.MOV)

    @classmethod
    def mp4(cls):
        '''
            Generate an MP4 Video file and return as bytes.
        '''
        return BinaryFile().image(file_type=VideoFile.MP4)


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .action import *

class HttpActionLoader:

    def __init__(self, endpoint):
        self._endpoint = endpoint

    def __getattr__(self, action: str, **fargs) -> HttpEndPointAction:
        if action == "_anon":
            msg = AnonEndPointAction(endpoint=vars(self)["_endpoint"])
        else:
            msg = HttpEndPointAction(name=action, endpoint=vars(self)["_endpoint"], **fargs)
        return msg

    def perform(self, **fargs):
        self._anon.perform(**fargs)


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class HttpCookie:

    def __init__(self, session, cookie):
        self.__cookie = cookie

    @property
    def _cookie(self):
        return self.__cookie

    @property
    def value(self):
        '''
        Value stored in this cookie.
        '''
        return self._cookie.value

    @property
    def secure(self):
        '''
        If secure flag is set for this cookie, this value is True, else False.
        '''
        return self._cookie.secure

    @property
    def httponly(self):
        '''
        If HttpOnly flag is set for this cookie, this value is True, else False.
        '''
        return self._cookie.has_nonstandard_attr("HttpOnly")



# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random

class Number:

    @classmethod
    def fixed_length_number(cls, *, length) -> int:
        '''
            Generate a fixed length number

            Keyword Arguments:
                length: Number of digits in generated number.

            Returns:
                A generated fixed length number

            Note:
                A number of minimum length 1 is always generated.
        '''
        arr0 = [str(random.randint(1,9))]
        arr = []
        if length >= 2:
            arr = [str(random.randint(0,9)) for i in range(length-1)]
        arr0.extend(arr)
        return int("".join(arr0))

    @classmethod
    def int(cls, *, end, begin=0) -> int:
        '''
            Generate a random integer.

            Keyword Arguments:
                end: (inclusive) upper limit for the integer
                begin: (inclusive) lower limit for the integer. Default is 0.

            Returns:
                A generated integer
        '''
        return random.randint(begin, end)



# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

##########################################################
# Raised and consumed by internal implementation of Arjuna
##########################################################

import re
from arjuna.tpi.data.entity import _DataEntity

def arj_convert(content):

        if content is None:
            return content

        if type(content) is str and content == "":
            return content

        from json import JSONEncoder, dumps, loads
        from arjuna.tpi.parser.yaml import YamlDict, YamlList
        from arjuna.tpi.parser.json import JsonDict, JsonList
        from arjuna.tpi.data.entity import _DataEntity
        from arjuna.tpi.helper.arjtype import CIStringDict

        class _CustomEncoder(JSONEncoder):
            def default(self, o):
                if isinstance(o, YamlDict) or isinstance(o, YamlList) or isinstance(o, JsonList) or isinstance(o, JsonDict):
                    return o.raw_object
                elif isinstance(o, _DataEntity):
                    return o.as_dict()
                elif isinstance(o, CIStringDict):
                    return o.store
                return JSONEncoder.default(self, o)

        if type(content) is tuple:
            content = list(content)

        # if content:
        if isinstance(content, JsonList) or isinstance(content, JsonDict) or isinstance(content, YamlList) or isinstance(content, YamlDict):
            content = content.raw_object
        elif isinstance(content, _DataEntity):
            content = content.as_dict()
        elif isinstance(content, CIStringDict):
            content = content.store
        elif content == "null":
            content = None
        else:
            if not content:
                if content is not None:
                    content = str(content)

        content = dumps(content, cls=_CustomEncoder, indent=2)

        if content != "":
            return loads(content)
        else:
            return ""


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid
from mimesis import locales as Locales
from mimesis import Text as MimesisText
from .number import Number

class Text:

    @classmethod
    def ustr(cls, *, prefix: str=None, maxlen: int=None, minlen: int=None, delim:str="-", strict=False) -> str:
        '''
            Generate a unique UUID string.
            If minlen/maxlen are specified in a manner that leads to uuid truncation, uniqueness is not enforced.
            Base string length is prefix length + delim length + 36 (length of uuid4)
            Different arguments tweak the length of generated string by appending uuid one or more times fully or partially.
            
            Keyword Arguments:
                prefix: (Optional) prefix to be added to the generated UUID string.
                minlen: (Optional) Minimum length of the retuned string (inclusive of prefix + delim). Default is base string length.
                maxlen: (Optional) Maximum length of the retuned string (inclusive of prefix + delim). Should be greater than minlen. Default is calculated as:
                    * if minlen > half of base string length, then default maxlen is 2 * minlen 
                    * if minlen <  half of base string length, then default maxlen is base string length

                delim: (Optional) Delimiter between prefix and generated string. Default is "-". Ignored if prefix is not specified.
                strict: (Optional) If True uniqueness of string is enforced which means full generated uuid must be used atleast once. This means length of generated string must be >= base string length, else an exception is thrown.
            Returns:
                A string that is unique for current session.
        '''
        if maxlen is not None and minlen is not None:
            if maxlen < minlen:
                raise Exception("maxlen must be greater than minlen for Random.ustr call")

        prefix = prefix and prefix + delim or ""
        gen_str = uuid.uuid4()
        base_str = prefix + str(gen_str)

        if strict:
            if minlen is not None and minlen < len(base_str):
                minlen = len(base_str)
            if maxlen is not None and maxlen < len(base_str):
                raise Exception("In strict mode the maxlen can not be less than length of prefix + delim + 36 (len of uuid). Generated string: >>{}<< (length={}). Specified maxlen: {}".format(base_str, len(base_str), maxlen))

        target_len = len(base_str)
        if minlen is not None and maxlen is not None:
            target_len = Number.int(begin=minlen, end=maxlen)
        elif minlen is not None:
            if minlen < 0.5 * len(base_str):
                target_len = Number.int(begin=minlen, end=len(base_str))
            else:
                target_len = Number.int(begin=minlen, end=2 * minlen)
        elif maxlen is not None:
            if maxlen <= len(base_str):
                target_len = maxlen
            else:
                target_len = Number.int(begin=len(base_str), end=maxlen)

        multiplier = (target_len // 36) + (target_len % 36 > 0 and 1 or 0)
        base_str = prefix + multiplier * str(gen_str)

        return base_str[:target_len]

    @classmethod
    def fixed_length_str(cls, *, length) -> str:
        '''
            Generate a fixed length string

            Keyword Arguments:
                length: Number of chracters in generated number.

            Returns:
                A generated fixed length string

            Note:
                A number of minimum length 1 is always generated.
        '''
        return cls.ustr(minlen=length, maxlen=length)[:length]

    @classmethod
    def sentence(cls, *, locale=Locales.EN) -> str:
        '''
            Generate a sentence

            Keyword Arguments:
                locale: (Optional) locale for generating sentence

            Returns:
                A generated sentence
        '''
        return MimesisText(locale).sentence()

    @classmethod
    def alphabet(cls, *, locale=Locales.EN, lower_case=False):
        '''
            Generate a random integer.

            Keyword Arguments:
                locale: (Optional) Locale for data
                lower_case: (inclusive) If true lower case alphabet is returned where applicable.
            Returns:
                A list of alphabet characters.
        '''
        return MimesisText(locale).alphabet(lower_case=lower_case)

    @classmethod
    def answer(cls, *, locale=Locales.EN) -> str:
        '''
            Generate a random answer.

            Keyword Arguments:
                locale: (Optional) Locale for data
        '''
        return MimesisText(locale).answer()

    @classmethod
    def level(cls, *, locale=Locales.EN) -> str:
        '''
            Generate a random level for danger etc.

            Keyword Arguments:
                locale: (Optional) Locale for data
        '''
        return MimesisText(locale).level()

    @classmethod
    def quote(cls, *, locale=Locales.EN) -> str:
        '''
            Generate a random quote.

            Keyword Arguments:
                locale: (Optional) Locale for data
        '''
        return MimesisText(locale).quote()

    @classmethod
    def word(cls, *, locale=Locales.EN) -> str:
        '''
            Generate a random word.

            Keyword Arguments:
                locale: (Optional) Locale for data
        '''
        return MimesisText(locale).word()

    @classmethod
    def swear_word(cls, *, locale=Locales.EN) -> str:
        '''
            Generate a random swear word.

            Keyword Arguments:
                locale: (Optional) Locale for data
        '''
        return MimesisText(locale).swear_word()

    @classmethod
    def words(cls, *, locale=Locales.EN, count=5): # -> list[str]: Readthedocs has max version 3.8, this can be used 3.9 onwards
        '''
            Generate a random word.

            Keyword Arguments:
                locale: (Optional) Locale for data
                count: Number of words. Default is 5.
        '''
        return MimesisText(locale).words(count=count)


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from mimesis import locales as Locales
from mimesis import Person as MimesisPerson

class Person:

    @classmethod
    def first_name(cls, *, locale=Locales.EN) -> str:
        '''
            Generate a first name.

            Keyword Arguments:
                locale: (Optional) locale for generating first name

            Returns:
                A generated first name
        '''
        return MimesisPerson(locale).first_name()

    @classmethod
    def last_name(cls, *, locale=Locales.EN) -> str:
        '''
            Generate a last name.

            Keyword Arguments:
                locale: (Optional) locale for generating last name

            Returns:
                A generated last name
        '''
        return MimesisPerson(locale).last_name()

    @classmethod
    def name(cls, *, locale=Locales.EN) -> str:
        '''
            Generate a full name (first name and last name).

            Keyword Arguments:
                locale: (Optional) locale for generating phone number

            Returns:
                A generated full name
        '''
        return "{} {}".format(cls.first_name(locale=locale), cls.last_name(locale=locale))

    @classmethod
    def phone(cls, *, locale=Locales.EN) -> str:
        '''
            Generate a phone number.

            Keyword Arguments:
                locale: (Optional) locale for generating phone number

            Returns:
                A generated phone number
        '''
        return MimesisPerson(locale).telephone()

    @classmethod
    def email(cls, *, locale=Locales.EN, name=None, domain=None) -> str:
        '''
            Generate an email address.

            Keyword Arguments:
                locale: (Optional) locale for generating email address
                name: (Optional) If provided, then email address is created as <provided name>@<generated domain>
                domain: (Optional) If provided, then email address is created as <generated name>@<provided domain>

            Returns:
                A generated email address
        '''
        gname, gdomain = MimesisPerson(locale).email().rsplit("@", 1)
        name = name is not None and name or gname
        domain = domain is not None and domain or gdomain
        return f"{name}@{domain}"


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from arjuna.tpi.httpauto.response import HttpResponse

class BaseHttpEndPoint:

    def __init__(self, *, name, service, ep_dir):
        self.__name = name
        self.__service = service
        self.__root_dir = ep_dir
        from .actionloader import HttpActionLoader
        self.__actions = HttpActionLoader(endpoint=self)

    @property
    def name(self):
        return self.__name

    @property
    def service(self):
        return self.__service

    @property
    def root_dir(self):
        return self.__root_dir

    @property
    def message(self):
        '''
            Http Message Loader for this service using default end point.
        '''
        return self.action._anon._message

    @property
    def action(self):
        '''
            Http Message Loader for this service end point.
        '''
        return self.__actions

    def send(self, msg=None, **fargs) -> HttpResponse:
        return self.action._anon.send(msg=msg, **fargs)

    def perform(self, action, **fargs) -> HttpResponse:
        if action is None:
            self.__actions._anon.perform(**fargs)
        else:
            getattr(self.__actions, action).perform(**fargs)

class AnonHttpEndPoint(BaseHttpEndPoint):

    def __init__(self, *, service):
        super().__init__(name="anon", service=service, ep_dir=service.root_dir)

class HttpEndPoint(BaseHttpEndPoint):
    def __init__(self, *, name, service):
        super().__init__(name=name, service=service, ep_dir=os.path.join(service.root_dir, name))




# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re

from arjuna.core.utils import data_utils
from arjuna.core.types import constants
from arjuna.tpi.data.entity import _DataEntity

class TextFile:
    '''
    Represents a text file in Read mode.

    Arguments:
        fpath: Absolute path of text file.
    '''

    def __init__(self, fpath):
        self._fpath = fpath
        self._f = open(self._fpath, "r")

    def close(self):
        '''
           Close the file handle.
        '''
        self._f.close()

    def read(self) -> str:
        '''
           Read contents of text file.

           Returns:
                Content of text file as a Python str object.
        '''
        return self._f.read()

    def _get_file(self):
        return self._f


class _TextResource(TextFile):
    def __init__(self, file_name):
        super().__init__(
        os.path.abspath(os.path.join(os.path.dirname(__file__), os.sep.join(["../..", "res", file_name]))))


class TextFileAsLines:
    '''
        Represents a text file in Read mode to get line by line content.

        Arguments:
            fpath: Absolute path of text file.

        Keyword Arguments:
            **formatters: Arbitrary key-value arguments to format a line before returning.

        Note:
            You can loop over this object:

                .. code-block:: python

                    for line in fobj:
                        # Do something about the content
                        print(line)
    '''

    def __init__(self, fpath, **formatters):
        self.__filereader = open(fpath, "r")
        self.__formatters = formatters

    def __iter__(self):
        return self

    def next(self) -> str:
        '''
            Get next line.

            Returns:
                Line as a Python str object.
        '''
        l = self.__filereader.readline()
        if not l:
            self.close()
            raise StopIteration()           
        elif l.strip().startswith('#'):
            return self.next()
        else:
            if self.__formatters:
                return l.rstrip().format(**self.__formatters)
            else:
                return l.rstrip()

    __next__ = next

    def close(self):
        '''
           Close the file handle.
        '''
        self.__filereader.close()


class DelimTextFileWithLineAsSeq:
    '''
        Represents a text file in Read mode to get line by line content split by defined delimiter.

        Arguments:
            fpath: Absolute path of text file.

        Keyword Arguments:
            delimiter: (Optional) Delimiter that separates different parts of line. Default is tab (\t)
            header_line_present: (Optional) If True, the first line is treated as a header line.

        Note:
            You can loop over this object:

                .. code-block:: python

                    for line_parts in fobj:
                        # Do something about the parts of a current line
                        print(line_parts)
    '''

    def __init__(self, fpath, *, delimiter="\t", header_line_present=True):
        self.__filereader = TextFileAsLines(fpath)
        self.delimiter = delimiter
        self.__headers = None
        self._populate_headers()

    def __iter__(self):
        return self

    def _populate_headers(self):
        try:
            self.__headers = self.next()
        except:
            raise Exception("Invalid input file data. Empty headers line.")

    def next(self) -> tuple:
        '''
            Get next line parsed based on delimiter as a tuple.

            Returns:
                Line as a Python tuple object.
        '''
        line = self.__filereader.next()
        return data_utils.split(line.rstrip(), self.delimiter)

    __next__ = next

    @property
    def headers(self) -> tuple:
        '''
            Get header line parsed based on delimiter as a tuple.

            Returns:
                Header line as a Python tuple object. None if header line is not present.
        '''
        return self.__headers

    def close(self):
        '''
           Close the file handle.
        '''
        self.__filereader.close()


class DelimTextFileWithLineAsMap:
    '''
        Represents a text file in Read mode to get line by line content split by defined delimiter.

        The first line is treated as a header line.

        Arguments:
            fpath: Absolute path of text file.

        Keyword Arguments:
            delimiter: (Optional) Delimiter that separates different parts of line. Default is tab (\t)

        Note:
            You can loop over this object:

                .. code-block:: python

                    for line_parts in fobj:
                        # Do something about the parts of a current line
                        print(line_parts)
    '''

    def __init__(self, fpath, *, delimiter="\t"):
        self.__filereader = DelimTextFileWithLineAsSeq(fpath, delimiter=delimiter)

    def __iter__(self):
        return self

    @property
    def headers(self) -> tuple:
        '''
            Get header line parsed based on delimiter as a tuple.

            Returns:
                Header line as a Python tuple object.
        '''
        return self.__filereader.headers

    def close(self):
        '''
           Close the file handle.
        '''
        self.__filereader.close()

    def next(self) -> dict:
        '''
            Get next line parsed based on delimiter converted into a Python dictionary based on header line.

            Returns:
                Line as a Python dict object.
        '''
        line_parts = self.__filereader.next()

        if len(self.headers) != len(line_parts):
            raise Exception(
                "Invalid input file data. Number of headers and data values do not match.{0}Headers:{1}{0}Data values:{2}{0}".format(
                    os.linesep, self.headers, line_parts))
        else:
            return dict(zip(self.headers, line_parts))

    __next__ = next

from arjuna.tpi.engine.asserter import AsserterMixIn

class Text(str, AsserterMixIn):
    '''
        Arjuna representation of a String.
        Provides factory methods for dealing with reading text file content in various forms.

        Arguments:
            text: Input string. If a non-string object is provided it str(text) is called.
            fargs: Arbitrary positional arguments for formatting the string.

        Keyword Arguments:
            fnargs: Arbitrary named arguments (keworg arguments) for formatting the string.
    '''
    def __init__(self, text, *fargs, **fnargs):
        str.__init__(text)
        AsserterMixIn.__init__(self)
        self.__content = text
        self.__do_eval = False
        if type(self.__content) is not str:
            self.__do_eval = True
            self.__content = str(self.__content)
        self.__fargs = fargs
        self.__fnargs = fnargs
        if self.__fargs or self.__fnargs:
            self.format(*self.__fargs, **self.__fnargs)

    def format(self, *fargs, **fnargs):
        '''
            Format the string using provided named and keyword arguments.

            Positional arguments are used to format $$ placeholders as per the order in which they are found.

            Keyword/named arguments are used to format $<name>$ placeholders. Order does not matter.

            Supports C/L/R queries as names of placeholders.

            The provided features are very advanced. Refer Arjuna documentation for use cases and examples.

            Arguments:
                fargs: Arbitrary positional arguments for formatting the string.

            Keyword Arguments:
                fnargs: Arbitrary named arguments (keworg arguments) for formatting the string.
        '''

        target = self.__content

        def get_global_value(in_str):
            from arjuna import C, L, R
            gtype, query = in_str.split(".", 1)
            gtype = gtype.upper()
            return locals()[gtype](query)

        pos_pattern = r"(\$\s*\$)"
        named_pattern = r"\$(\s*[\w\.]+?\s*)\$"
        fmt_target = target.replace("{", "__LB__").replace("}", "__RB__")

        # Find params
        pos_matches = re.findall(pos_pattern, fmt_target)
        named_matches = re.findall(named_pattern, fmt_target)

        if pos_matches and named_matches:
            for match in named_matches:
                match = match.lower().strip()
                if not match.startswith("c.") and not match.startswith("l.") and not match.startswith("r."):
                    raise Exception("You can not use positional $$ placeholders and named $<name>$ placholders together in a withx locator definition (except those with C./L./R. prefixes to fetch global values. Wrong withx definition: " + target)

        if pos_matches:
            if not fargs:
                fargs = tuple()    
            if len(pos_matches) != len(fargs):
                raise Exception("Number of positional arguments supplied to format withx locator do not match number of $$ placeholders. Placeholders: {}. Positional args: {}. Wrong usage of withx locator: {}".format(len(pos_matches), fargs, target))

            for i,match in enumerate(pos_matches):
                fmt_target = fmt_target.replace(match, "{}")

            fmt_target = fmt_target.format(*fargs)

        def get_value_from_container(name, *, container=None):
            if container is None:
                parts = name.split(".", 1)
                if len(parts) == 1:
                    # It is reference directly to an object and if present could have been resolved before this logic hits.
                    return "__NOTFOUND"
                container, name = parts

            name_parts = re.split(r'([\.\[])', name, 1) # Will return single element list or 3 part list: name, delim, rest
            if len(name_parts) == 1:
                name, delim, rest = name, "", ""
            else:
                name, delim, rest = name_parts

            try:
                container = fnargs[container]
                if type(container) is dict:
                    first_obj = container[name]
                elif isinstance(type(container), _DataEntity):
                    first_obj = getattr(container, name)
                if delim == "":
                    return first_obj
                else:
                    return eval(f"first_obj{delim}{rest}")
            except Exception as e:
                return "__NOTFOUND"                 
            
        for match in named_matches:
            names_set = None
            target = "${}$".format(match)
            processed_name = match.lower().strip()
            repl_value = None
            if processed_name.startswith("c.") or processed_name.startswith("l.") or processed_name.startswith("r."):
                repl_value = get_global_value(processed_name)
            elif processed_name.startswith("data."):
                processed_name = processed_name.split(".")[1]
                if 'data' in fnargs:
                    temp_val = get_value_from_container(processed_name, container="data")
                    if temp_val == "__NOTFOUND":
                        continue
                    else:
                        repl_value = temp_val
                else:
                    continue
            else:
                if processed_name not in fnargs:
                    # Try a container and dynamic logic like a.b.c or a.b[3]
                    temp_val = get_value_from_container(processed_name)
                    if temp_val != "__NOTFOUND":
                        repl_value = temp_val
                    else:
                        if 'data' in fnargs:
                            temp_val = get_value_from_container(processed_name, container="data")
                            if temp_val == "__NOTFOUND":
                                continue
                            else:
                                repl_value = temp_val
                        else:
                            continue
                else:
                    repl_value = fnargs[processed_name]

            fmt_target = fmt_target.replace(target, str(repl_value))

        fmt_target = fmt_target.replace("__LB__", "{").replace("__RB__", "}")
        if self._eval:
            fmt_target = eval(fmt_target)

        self.__content = fmt_target
        return self.__content

    @property
    def content(self):
        return self.__content

    @property
    def _eval(self):
        return self.__do_eval

    def assert_contains(self, sub_str, *, msg):
        self.asserter.assert_true(sub_str in self.content, msg="Sub string {} not found in {}. ".format(sub_str, self.content) + msg)

    def findall(self, repattern):
        return re.findall(repattern, self.content)

    def find(self, repattern):
        all = re.findall(repattern, self.content)
        if all:
            return all[0]
        else:
            raise Exception(f"No match found for {repattern} in {self.content}")

    def exists(self, repattern):
        matches = re.findall(repattern, self.content)
        if matches:
            return True
        else:
            return False

    def __str__(self):
        return self.content

    def __repr__(self):
        return self.content

    def __eq__(self, other):
        if isinstance(other, Text):
            return self.content == other.content
        else:
            return self.content == other

    @classmethod
    def file_content(cls, fpath) -> str:
        '''
            Get content of a text file.

            Arguments:
                fpath: Absolute path of text file.

            Returns:
                Content of text file as Python str object.
        '''
        f = TextFile(fpath)
        content = f.read()
        f.close()
        return content

    @classmethod
    def file_lines(cls, fpath) -> TextFileAsLines:
        '''
            Get a reader for text file in Read mode to get line by line content.

            Arguments:
                fpath: Absolute path of text file.

            Returns:
                Arjuna's TextFileAsLines object.

            Note:
                You must explicitly close the returned object by calling its **close()** method.
        '''
        return TextFileAsLines(fpath)

    @classmethod
    def delimited_file(cls, fpath, *, delimiter="\t", header_line_present=True) -> 'DelimTextFileWithLineAsMap OR DelimTextFileWithLineAsSeq':
        '''
            Represents a text file in Read mode to get line by line content split by defined delimiter.

            Arguments:
                fpath: Absolute path of text file.

            Keyword Arguments:
                delimiter: (Optional) Delimiter that separates different parts of line. Default is tab (\t)
                header_line_present: (Optional) If True, the DelimTextFileWithLineAsMap is created, else DelimTextFileWithLineAsSeq is created to represent the file.

            Note:
                You can loop over the returned object:

                    .. code-block:: python

                        for line in fobj:
                            # Do something about the parts of a current line
                            print(line)

                The line is a tuple of parts for DelimTextFileWithLineAsSeq object and is a dictionary for DelimTextFileWithLineAsMap object.

            Note:
                You must explicitly close the returned object by calling its **close()** method.
        '''
        if header_line_present:
            return DelimTextFileWithLineAsMap(fpath, delimiter=delimiter)
        else:
            return DelimTextFileWithLineAsSeq(fpath, delimiter=delimiter)



# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from mimesis import locales

'''
Data Generator

This module contains classes and functions that generate data.
'''

from arjuna.engine.data.generator.text import Text
from arjuna.engine.data.generator.number import Number
from arjuna.engine.data.generator.person import Person
from arjuna.engine.data.generator.address import Address
from arjuna.engine.data.generator.entity import Entity
from arjuna.engine.data.generator.color import Color
from arjuna.engine.data.generator.file import File

class _gen:
    pass

class processor:
    '''
        Processes data by calling the provided callable with provided arguments.

        Arguments:
            callable: The callable to be called in processing.
            *args: Arbitrary positional args to be passed to the callable.

        Keyword Arguments:
            **kwargs: Arbitrary keyword arguments to tbe passed to the callable.
    '''

    def __init__(self, callable, *args, **kwargs):
        self.__callble = callable
        self.__args = args
        self.__kwargs = kwargs

    def process(self, data):
        '''
            Processes data in the provided data object by calling the callable with arguments provided in constructor.

            Arguments:
                data: Data object
        '''
        if type(self.__callble) is str:
            return getattr(data, self.__callble)(*self.__args, **self.__kwargs)
        else:
            return self.__callble(data_iterator, *self.__args, **self.__kwargs)  

class composer:
    '''
        Combines data in an iterable by calling the provided callable with provided arguments.

        Arguments:
            callable: The callable to be called in composing.
            *args: Arbitrary positional args to be passed to the callable.

        Keyword Arguments:
            **kwargs: Arbitrary keyword arguments to tbe passed to the callable.
    '''

    def __init__(self, callable, *args, **kwargs):
        self.__callble = callable
        self.__args = args
        self.__kwargs = kwargs

    def compose(self, data_iter):
        '''
            Combine data in the provided iterable by calling the callable with arguments provided in constructor.

            Arguments:
                data_iter: Data iterable
        '''
        return self.__callble(data_iter, *self.__args, **self.__kwargs)  

def _same(in_data):
    return in_data

class generator(_gen):
    '''
        Generate data by calling the provided callable with provided arguments.

        Arguments:
            callable: The callable to be called in composing. By default, same generated object is returned.
            *args: Arbitrary positional args to be passed to the callable.

        Keyword Arguments:
            processor: This callable is called after data is generated by passing generated data as argument. Useful for data transformation of any kind. If its type is string, it is assumed to be a method of the data which the data callable generated.
            **kwargs: Arbitrary keyword arguments to tbe passed to the callable.
    '''

    def __init__(self, callable, *args, processor=_same, **kwargs):
        self.__callble = callable
        self.__args = args
        self.__kwargs = kwargs
        self.__processor = processor

    def generate(self):
        '''
            Generate data by calling the callable with arguments provided in constructor.

            After generation the coverter callable is called with this data before returning the data.
        '''
        data = self.__callble(*self.__args, **self.__kwargs)
        if type(self.__processor) is str:
            return getattr(data, self.__processor)()
        elif isinstance(self.__processor, processor):
            return self.__processor.process(data)
        else:
            return self.__processor(data)

class composite(_gen):
    '''
        Composite Data Generator.

        Generate data by composing the output of all generators, callables or static data.

        Arguments:
            *generators_or_data: Arbitrary generators, functions or static data objects.

        Keyword Arguments:
            composer: This callable is called after data sequence is generated by passing generated data sequence as argument. Useful for data transformation of any kind.
    '''

    def __init__(self, *generators_or_data, composer=_same):
        self.__generators_or_data = generators_or_data
        self.__composer = composer

    def generate(self):
        '''
            Generate data by composing the output of all generators, callables or static data.
            
            The composer callable is called after data sequence is generated by passing generated data sequence as argument. Useful for data transformation of any kind.
        '''
        out = list()
        for gen_or_data in self.__generators_or_data:
            if hasattr(gen_or_data, '__call__'):
                out.append(gen_or_data())
            elif isinstance(gen_or_data, generator):
                out.append(gen_or_data.generate())
            else:
                out.append(gen_or_data)

        if isinstance(self.__composer, composer):
            return self.__composer.compose(out)
        else:
            return self.__composer(out)

class DataLocale:
    '''
        Data Locale object.

        When this object is passed where relevant to Random class methods, it localizes the data generated.

        To create a specific locale object, you can use the dot notation.

        .. code-block:: python

            locale.en
            locale.fr

        For list of supported locales, check :ref:`suppored_data_locales`. You also use the following code to get all supported locale names:

        .. code-block:: python

            locale.supported
    '''

    __SUPPORTED = {
        'cs': 'Czech', 
        'da': 'Danish', 
        'de': 'German', 
        'de_at': 'Austrian German', 
        'de_ch': 'Swiss German', 
        'el': 'Greek', 
        'en': 'English', 
        'en_gb': 'British English', 
        'en_au': 'Australian English', 
        'en_ca': 'Canadian English', 
        'es': 'Spanish', 
        'es_mx': 'Mexican Spanish', 
        'et': 'Estonian', 
        'fa': 'Farsi', 
        'fi': 'Finnish', 
        'fr': 'French', 
        'hu': 'Hungarian', 
        'is': 'Icelandic', 
        'it': 'Italian', 
        'ja': 'Japanese', 
        'kk': 'Kazakh', 
        'ko': 'Korean', 
        'nl': 'Dutch', 
        'nl_be': 'Belgium Dutch', 
        'no': 'Norwegian', 
        'pl': 'Polish', 
        'pt': 'Portuguese', 
        'pt_br': 'Brazilian Portuguese', 
        'ru': 'Russian', 
        'sv': 'Swedish', 
        'tr': 'Turkish', 
        'uk': 'Ukrainian', 
        'zh': 'Chinese'
    }

    _SINGLE_OBJ = {

    }

    def __getattr__(self, name):
        if name.lower() in self.__SUPPORTED:
            if name not in self._SINGLE_OBJ:
                self._SINGLE_OBJ[name] = getattr(locales, name.upper())
            return self._SINGLE_OBJ[name]
        else:
            raise Exception(f"{name} is not a support locale name for Random class methods. Allowed: {self.__SUPPORTED}")

    @property
    def supported(self):
        import copy
        return copy.deepcopy(self.__SUPPORTED)


class __RandomMetaClass(type):

    @property
    def locale(cls):
        return DataLocale()


class Random(metaclass=__RandomMetaClass):
    '''
        Provides methods to create random strings and numbers of different kinds.
    '''

    # Text
    ustr = Text.ustr
    fixed_length_str = Text.fixed_length_str
    sentence = Text.sentence
    alphabet = Text.alphabet

    # Number
    int = Number.int
    fixed_length_number = Number.fixed_length_number

    # Person
    first_name = Person.first_name
    last_name = Person.last_name
    name = Person.name
    phone = Person.phone
    email = Person.email

    # Address
    street_number = Address.street_number
    street_name = Address.street_name
    house_number = Address.house_number
    postal_code = Address.postal_code
    city = Address.city
    country = Address.country

    # Color
    color = Color.color
    rgb_color = Color.rgb_color
    hex_color = Color.hex_color

    # # File
    # mp3 = File.mp3
    # aac = File.aac
    # zip = File.zip
    # gzip = File.gzip
    # pdf = File.pdf
    # docx = File.docx
    # xlsx = File.xlsx
    # pptx = File.pptx
    # gif = File.gif
    # png = File.png
    # jpg = File.jpg
    # mov = File.mov
    # mp4 = File.mp4

    # Entities
    person = Entity.person
    address = Entity.address




        


    


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from arjuna.tpi.helper.arjtype import NotFound

class TextExtractor:

    def __init__(self, response):
        self.__response =  response

    @property
    def response(self):
        return self.__response

    def store(self, name, regex, strict):
        try:
            value = self.response.text.find(regex)
        except Exception:
            if not strict:
                value = NotFound()
            else:
                raise Exception(f"Issue in extracting value for >{name}< as no element was found using regex >{regex}<.")
        
        self.response.store[name] = value




# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from arjuna.tpi.data.entity import data_entity

Person = data_entity(
        "Person", 
        qualification=None,
        age=None,
        blood_type=None,
        email=None,
        first_name=None,
        last_name=None,
        name=None,
        gender=None,
        height=None,
        id=None,
        language=None,
        nationality=None,
        occupation=None,
        phone=None,
        title=None,
        university=None,
        weight=None,
        work_experience=None
    )